#!/bin/bash

# spamtrap.sh -- a script for blacklisting spammers
# Copyright (C) 2013 Sam Clippinger (samc (at) silence (dot) org)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


SERVER_NAME=iconoclast.silence.org
TOP_MAIL_FOLDER=/home/vpopmail/domains
CATCHALL_JUNK_FOLDER=${TOP_MAIL_FOLDER}/silence.org/spamtrap/Maildir/.Junk
TMP_FOLDER=/home/vpopmail/traps/tmp
IP_BLACKLIST_FILE=/home/vpopmail/blacklist_ip
UNBLOCKABLE_RDNS_FILE=/home/vpopmail/unblacklistable_domains
UNBLOCKABLE_IP_FILE=/etc/rc.d/unblockable_ips
LOG_FILE=/home/vpopmail/traps/spamtrap/spamtrap.log
FILE_OWNER=vpopmail.vchkpw

export DAYS_TO_SAVE=5
export DAYS_TO_CONSIDER=3
export JUNK_SPAMS_REQUIRED=3
export SPAMTRAP_SPAMS_REQUIRED=1
export GUILTY_BY_ASSOCIATION=1

rnd=${RANDOM}
tmp_ip_blacklist_file=${TMP_FOLDER}/spamtrap.blacklist_ip.${rnd}
tmp_junk_grep_file=${TMP_FOLDER}/spamtrap.junk.grep.${rnd}
tmp_spamtrap_grep_file=${TMP_FOLDER}/spamtrap.spamtrap.grep.${rnd}
tmp_good_grep_file=${TMP_FOLDER}/spamtrap.good.grep.${rnd}
tmp_junk_raw_file=${TMP_FOLDER}/spamtrap.junk.raw.${rnd}
tmp_junk_ip_file=${TMP_FOLDER}/spamtrap.junk.ip.${rnd}
tmp_spamtrap_raw_file=${TMP_FOLDER}/spamtrap.spamtrap.raw.${rnd}
tmp_spamtrap_ip_file=${TMP_FOLDER}/spamtrap.spamtrap.ip.${rnd}
tmp_assoc_junk_ip_file=${TMP_FOLDER}/spamtrap.assoc.junk.ip.${rnd}
tmp_good_ip_file=${TMP_FOLDER}/spamtrap.good.ip.${rnd}
lock_file=${TMP_FOLDER}/spamtrap.lock

if [ ! -f ${lock_file} ]
then
  touch ${lock_file}

  for file in `find ${CATCHALL_JUNK_FOLDER} -mmin +$[${DAYS_TO_SAVE}*24*60] -type f -print`
  do
    rm -f ${file}
  done

  for file in `grep -lir "from: postmaster@" ${CATCHALL_FOLDER}`
  do
    rm -f ${file}
  done

  for file in `grep -lir "from: mailer-daemon@" ${CATCHALL_FOLDER}`
  do
    rm -f ${file}
  done

  touch ${tmp_junk_grep_file} ${tmp_good_grep_file} ${tmp_junk_ip_file} ${tmp_junk_raw_file} ${tmp_assoc_junk_ip_file} ${tmp_good_ip_file} ${tmp_ip_blacklist_file} ${tmp_spamtrap_grep_file} ${tmp_spamtrap_raw_file} ${tmp_spamtrap_ip_file}

  find ${TOP_MAIL_FOLDER} -type f \( -wholename */cur/* -o -wholename */new/* \) ! -wholename */.Junk/* ! -wholename */.Trash/* ! -wholename */.Sent/* -print0 | xargs -0 grep -h -m 1 ${SERVER_NAME} 2> /dev/null >> ${tmp_good_grep_file}
  find ${TOP_MAIL_FOLDER} -type f -mmin -$[1440*${DAYS_TO_CONSIDER}] \( -wholename */.Junk/cur/* -o -wholename */.Junk/new/* \) ! -wholename */silence.org/spamtrap/* -print0 | xargs -0 grep -h -m 1 ${SERVER_NAME} 2> /dev/null >> ${tmp_junk_grep_file}
  find ${CATCHALL_FOLDER} -type f -mmin -$[1440*${DAYS_TO_CONSIDER}] -print0 | xargs -0 grep -h -m 1 ${SERVER_NAME} 2> /dev/null >> ${tmp_spamtrap_grep_file}

  grep Received: ${tmp_junk_grep_file} | grep qmail-scanner | awk '{ print $3 }' | sort > ${tmp_junk_raw_file}
  cat ${tmp_junk_raw_file} | uniq -c | awk '{ if ($1 >= ENVIRON["JUNK_SPAMS_REQUIRED"]) print $2 }' > ${tmp_junk_ip_file}
  grep Received: ${tmp_spamtrap_grep_file} | grep qmail-scanner | awk '{ print $3 }' | sort > ${tmp_spamtrap_raw_file}
  cat ${tmp_spamtrap_raw_file} | uniq -c | awk '{ if ($1 >= ENVIRON["SPAMTRAP_SPAMS_REQUIRED"]) print $2 }' > ${tmp_spamtrap_ip_file}
  grep Received: ${tmp_good_grep_file} | grep qmail-scanner | awk '{ print $3 }' | sort | uniq > ${tmp_good_ip_file}

  for ip in `cat ${tmp_junk_ip_file} ${tmp_spamtrap_ip_file}`
  do
    class_c=`echo ${ip} | sed -e "s/\./ /g" | awk '{ print $1 "." $2 "." $3 "." }'`
    for new_ip in `grep -h ${class_c} ${tmp_junk_raw_file} ${tmp_spamtrap_raw_file} | sort | uniq`
    do
      if [ `grep -h ${new_ip} ${tmp_junk_raw_file} ${tmp_spamtrap_raw_file} | wc -l` -ge ${GUILTY_BY_ASSOCIATION} ]
      then
        echo ${new_ip} >> ${tmp_assoc_junk_ip_file}
      fi
    done
  done

  for ip in `cat ${tmp_assoc_junk_ip_file} | sort | uniq`
  do
    output=`grep ${ip} ${tmp_good_grep_file}`
    if [ -z "${output}" ]
    then
      doblock="true"
      for unblockable in `cat ${UNBLOCKABLE_IP_FILE}`
      do
        if [ "${ip:0:${#unblockable}}" == "${unblockable:0:${#ip}}" ]
        then
          doblock=""
        fi
      done

      tmp_name=`dnsptr ${ip}`
      if [ ! -z "${tmp_name}" ]
      then
        tmp_base=`domainsplit ${tmp_name}`
        output=`grep ${tmp_base} ${UNBLOCKABLE_RDNS_FILE}`
        if [ ! -z "${output}" ]
        then
          doblock=""
        fi
      fi

      if [ -n "${doblock}" ]
      then
        echo ${ip} >> ${tmp_ip_blacklist_file}
      fi
    fi
  done

  raw_junk_ips=`cat ${tmp_junk_ip_file} | wc -l`
  raw_spamtrap_ips=`cat ${tmp_spamtrap_ip_file} | wc -l`
  raw_good_ips=`cat ${tmp_good_ip_file} | wc -l`
  blacklisted_ips=`cat ${tmp_ip_blacklist_file} | wc -l`
  echo `date` Found ${raw_junk_ips} junk IPs, ${raw_spamtrap_ips} spamtrap IPs, ${raw_good_ips} good IPs and kept ${blacklisted_ips} >> ${LOG_FILE}

  chown ${FILE_OWNER} ${tmp_ip_blacklist_file}
  mv ${tmp_ip_blacklist_file} ${IP_BLACKLIST_FILE}
  rm -f ${tmp_junk_grep_file} ${tmp_good_grep_file} ${tmp_junk_ip_file} ${tmp_junk_raw_file} ${tmp_assoc_junk_ip_file} ${tmp_good_ip_file} ${tmp_ip_blacklist_file} ${tmp_spamtrap_grep_file} ${tmp_spamtrap_raw_file} ${tmp_spamtrap_ip_file}

  rm -f ${lock_file}
fi
